@@ -220,7 +220,7 @@ class H2OXGBoostMOJOModel(H2OTreeBasedSupervisedMOJOModelParams, HasMonotoneCons
 
     def getGpuId(self):
         value = self._java_obj.getGpuId()
-        return value
+        return H2OTypeConverters.scalaArrayToPythonArray(value)
 
 
     def getInteractionConstraints(self):
@@ -321,3 +321,8 @@ class H2OXGBoostMOJOModel(H2OTreeBasedSupervisedMOJOModelParams, HasMonotoneCons
     def getExportCheckpointsDir(self):
         value = self._java_obj.getExportCheckpointsDir()
         return value
+
+
+    def getAucType(self):
+        value = self._java_obj.getAucType()
+        return value
